/*
  Copyright (C) 2012 Jon Macey

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received m_a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "MyGLWindow.h"
#include <ngl/Colour.h>
#include <ngl/Mat4.h>
#include <ngl/Transformation.h>
#include <ngl/TransformStack.h>
#include <ngl/Material.h>
#include <ngl/VAOPrimitives.h>
#include <ngl/ShaderLib.h>

#include <iostream>

MyGLWindow::MyGLWindow(EGLconfig *_config) : EGLWindow(_config)
{
	std::cout<<"My GL Window Ctor\n";
	makeSurface();
}
MyGLWindow::~MyGLWindow()
{

}

void MyGLWindow::initializeGL()
{
	GLuint renderbuffers[2];

	// Let's create multiple names/ids at once.
	// To do this we declared our variable as a pointer *renderbuffers.
	glGenRenderbuffers(2, renderbuffers);

	// The index 0 will be our color render buffer.
	glBindRenderbuffer(GL_RENDERBUFFER, renderbuffers[0]);
	glRenderbufferStorage(GL_RENDERBUFFER, GL_RGBA4, m_width, m_height);
	glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_RENDERBUFFER, renderbuffers[0]);

	// The index 1 will be our depth render buffer.
	glBindRenderbuffer(GL_RENDERBUFFER, renderbuffers[1]);
	glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT16, m_width, m_height);
	glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, renderbuffers[1]);

	glBindRenderbuffer(GL_RENDERBUFFER, renderbuffers[0]);
	// Set background color and clear buffers
	glClearColor(0.2, 0.2,0.2,1.0);
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	// Enable back face culling.
  glEnable(GL_DEPTH_TEST);
	// now to load the shader and set the values
	// grab an instance of shader manager
	ngl::ShaderLib *shader=ngl::ShaderLib::instance();
	// we are creating a shader called Phong
	shader->createShaderProgram("Phong");
	// now we are going to create empty shaders for Frag and Vert
	shader->attachShader("PhongVertex",ngl::VERTEX);
	shader->attachShader("PhongFragment",ngl::FRAGMENT);
	// attach the source
	shader->loadShaderSource("PhongVertex","shaders/Vertex.vs");
	shader->loadShaderSource("PhongFragment","shaders/Fragment.fs");
	// compile the shaders
	shader->compileShader("PhongVertex");
	shader->compileShader("PhongFragment");
	// add them to the program
	shader->attachShaderToProgram("Phong","PhongVertex");
	shader->attachShaderToProgram("Phong","PhongFragment");
	// now bind the shader attributes for most NGL primitives we use the following
	// layout attribute 0 is the vertex data (x,y,z)
	shader->bindAttribute("Phong",0,"inVert");
	// attribute 1 is the UV data u,v (if present)
	shader->bindAttribute("Phong",1,"inUV");
	shader->bindAttribute("Phong",2,"inNormal");
  m_location=shader->getAttribLocation("Phong","tex");

	// now we have associated this data we can link the shader
	shader->linkProgramObject("Phong");

	(*shader)["Phong"]->use();

	// Now we will create a basic Camera from the graphics library
	// This is a static camera so it only needs to be set once
	// First create Values for the camera position
	ngl::Vec4 From(0,1,4);
	ngl::Vec4 To(0,0,0);
	ngl::Vec4 Up(0,1,0);
	// now load to our new camera
	m_cam= new ngl::Camera(From,To,Up,ngl::PERSPECTIVE);
	// set the shape using FOV 45 Aspect Ratio based on Width and Height
	// The final two are near and far clipping planes of 0.5 and 10
	m_cam->setShape(45,(float)720.0/576.0,0.05,350,ngl::PERSPECTIVE);
	// now create our light this is done after the camera so we can pass the
	// transpose of the projection Mat4 to the light to do correct eye space
	// transformations
 // first we create a mesh from an obj passing in the obj file and texture
	std::cout<<"building mesh\n";
  m_mesh = new ngl::Obj("models/Helix.obj","textures/helix_base.tif");
	 std::cout<<"create vao\n";
   // now we need to create this as a VAO so we can draw it
	 m_mesh->createVAO();
}


void MyGLWindow::loadMatricesToShader(
																			 ngl::TransformStack &_tx
																		 )
{
	ngl::ShaderLib *shader=ngl::ShaderLib::instance();
	(*shader)["Phong"]->use();

  ngl::Mat4 MVP=_tx.getCurrAndGlobal().getMatrix()*m_cam->getVPMatrix();
  shader->setShaderParamFromMatrix("MVP",MVP);

}



void MyGLWindow::paintGL()
{
	// Start with a clear screen
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	ngl::ShaderLib *shader=ngl::ShaderLib::instance();
	(*shader)["Phong"]->use();

	// Rotation based on the mouse position for our global transform
	ngl::Transformation trans;
	ngl::Mat4 rotX;
	ngl::Mat4 rotY;
	// create the rotation matrices
	static float x=0;
	static float y=0;
	// multiply the rotations
	ngl::Mat4 final=rotY*rotX;
	++x;
	++y;
	// add the translations
	// set this in the TX stack
	trans.setMatrix(final);
	m_transformStack.setGlobal(trans);
	// draw
	m_transformStack.setRotation(0,y,0);
	m_transformStack.setPosition(0,-0.5,0);
	m_transformStack.setScale(0.8,0.8,0.8);
	loadMatricesToShader(m_transformStack);
  // draw the mesh
  m_mesh->draw();


	glFlush();
	glFinish();
	swapBuffers();
}
